/*
	rpipic - A PIC Programmer for the Raspberry Pi
	Copyright (C) 2012  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/


#include "data.hpp"

#include <cstring>

uint8_t Data::GetData(uint32_t address)
{
	uint8_t byte = 0;
	if (!GetData(address, byte)){
		throw 1;
	}
	return byte;
}

bool Hex::GetData(uint32_t address, uint8_t &byte)
{
	for (auto it = items.begin(); it != items.end(); it++){
		auto &item = *it;
		if ((item.address <= address) && (address - item.address < item.bytes.size())){
			byte = item.bytes[address - item.address];
			return true;
		}
	}
	return false;
}

void Hex::SetData(uint32_t address, uint8_t byte)
{
	for (auto it = items.begin(); it != items.end(); it++){
		HexDataItem &item = *it;
		if (item.address <= address){
			if (address - item.address < item.bytes.size()){
				item.bytes[address - item.address] = byte;
				return;
			}
			else if (item.bytes.size() == address - item.address){
				item.bytes.push_back(byte);
				return;
			}
		}
	}
	HexDataItem item;
	item.address = address;
	item.bytes.push_back(byte);
	items.push_back(item);
}

void Hex::ReadFile(FILE *f)
{
	uint32_t baseaddress = 0;
	char line[524]; //a line should never be longer than 523 characters
	char c = 0;
	unsigned int linen = 1;
	size_t linel = 0;
	while (1 == fread(&c, 1, 1, f)){
		if ('\r' == c){
			continue;
		}
		if ('\n' == c){
			//process the line
			if (!linel){
				continue;
			}
			if (linel < 10){
				fprintf(stderr, "Line %i is incomplete: \"%s\"\n", linen, line);
				throw 0;
			}
			
			if (':' != line[0]){
				fprintf(stderr, "Line %i did not begin with ':': \"%s\"\n", linen, line);
				throw 0;
			}
			
			uint8_t length = ConvertHex(line[1], line[2]);
			uint16_t address = ConvertHex(line[5], line[6]) | (ConvertHex(line[3], line[4]) << 8);
			uint8_t type = ConvertHex(line[7], line[8]);
			
			if (linel != (unsigned int)length*2 + 11){
				fprintf(stderr, "Line %i is the wrong length: \"%s\"\n", linen, line);
				throw 0;
			}
			
			uint8_t checksum = 0;
			checksum += length;
			checksum += address >> 8;
			checksum += address & 0xFF;
			checksum += type;
			
			for (uint8_t i = 0; i < length; i++){
				checksum += ConvertHex(line[i*2 + 9], line[i*2 + 10]);
			}
			
			switch (type)
			{
				case 0:
					for (uint8_t i = 0; i < length; i++){
						SetData(baseaddress + address + i, ConvertHex(line[i*2 + 9], line[i*2 + 10]));
					}
					break;
				case 1:
					break;
				case 2:
					fprintf(stderr, "Line %i has unexpected type: \"%s\"\n", linen, line);
					throw 0;
				case 3:
					fprintf(stderr, "Line %i has unexpected type: \"%s\"\n", linen, line);
					throw 0;
				case 4:
					if (length != 2){
						fprintf(stderr, "Line %i has unexpected length: \"%s\"\n", linen, line);
						throw 0;
					}
					baseaddress =
						(ConvertHex(line[ 9], line[10]) << 16) |
						(ConvertHex(line[11], line[12]) << 24);
					break;
				case 5:
					fprintf(stderr, "Line %i has unexpected type: \"%s\"\n", linen, line);
					throw 0;
				default:
					fprintf(stderr, "Line %i has unknown type: \"%s\"\n", linen, line);
					throw 0;
			}
			
			checksum = -checksum;
			if (checksum != ConvertHex(line[length*2 + 9], line[length*2 + 10])){
				fprintf(stderr, "Line %i checksum incorrect (0x%02X): \"%s\"\n", linen, checksum, line);
				throw 0;
			}
			
			if (1 == type){
				break;
			}
			
			line[0] = 0;
			linel = 0;
			linen++;
		}
		else {
			if(linel >= 523){
				fprintf(stderr, "Line %i checksum too long: \"%s\"\n", linen, line);
				throw 0;
			}
			line[linel] = c;
			linel++;
			line[linel] = 0;
		}
	}
}

void Hex::WriteFile(FILE *f)
{
	uint32_t baseaddress = 0;
	for (auto it = items.begin(); it != items.end(); it++){
		auto &item = *it;
		for (unsigned int i = 0; i < item.bytes.size();){
			if (((item.address + i)  & 0xFFFF0000) != baseaddress){
				baseaddress = (item.address + i) & 0xFFFF0000;
				uint8_t checksum = 6;
				checksum += (baseaddress >> 16);
				checksum += (baseaddress >> 24);
				checksum = -checksum;
				fprintf(f, ":02000004%04X%02X\n", baseaddress >> 16, checksum);
			}
			unsigned int length = item.bytes.size() - i;
			if (length > 0x10){
				length = 0x10;
			}
			
			uint16_t hexaddr = (item.address + i) & 0xFFFF;
			fprintf(f, ":%02X%04X00", length, hexaddr);
			uint8_t checksum = 0;
			checksum += length >> 8;
			checksum += length & 0xFF;
			checksum += hexaddr >> 8;
			checksum += hexaddr & 0xFF;
			for (unsigned int j = 0; j < length; j++){
				fprintf(f, "%02X", item.bytes[i + j]);
				checksum += item.bytes[i + j];
			}
			checksum = -checksum;
			fprintf(f, "%02X\n", checksum);
			i += length;
		}
	}
	fprintf(f, ":00000001FF\n");
}

uint8_t Hex::ConvertHex(const char a, const char b)
{
	uint8_t n = 0;
	if (a < 0x30){
		fprintf(stderr, "Character '%c' is not hex\n", a);
		throw 0;
	}
	else if (a < 0x3A){
		n = (a - 0x30) << 4;
	}
	else if (a < 0x41){
		fprintf(stderr, "Character '%c' is not hex\n", a);
		throw 0;
	}
	else if (a < 0x47){
		n = (a - (0x41 - 10)) << 4;
	}
	else {
		fprintf(stderr, "Character '%c' is not hex\n", a);
		throw 0;
	}
	if (b < 0x30){
		fprintf(stderr, "Character '%c' is not hex\n", b);
		throw 0;
	}
	else if (b < 0x3A){
		n |= (b - 0x30);
	}
	else if (b < 0x41){
		fprintf(stderr, "Character '%c' is not hex\n", b);
		throw 0;
	}
	else if (b < 0x47){
		n |= (b - (0x41 - 10));
	}
	else {
		fprintf(stderr, "Character '%c' is not hex\n", b);
		throw 0;
	}
	return n;
}

