/*
	rpipic - A PIC Programmer for the Raspberry Pi
	Copyright (C) 2012  Nicholas Tomlinson, Gert van Loo, Dom Cobley
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/


#include <cstdio>
#include <ctime>

#include "gpio.hpp"

void FileGpio::Enable(int n, bool out)
{
	char filename[256];
	char value[256];
	snprintf(value, 256, "%i", n);
	snprintf(filename, 256, "/sys/class/gpio/gpio%i/direction", n);
	WriteFile("/sys/class/gpio/export", value);
	WriteFile(filename, out ? "out" : "in");
}

void FileGpio::Disable(int n)
{
	char value[256];
	snprintf(value, 256, "%i", n);
	Write(n, false);
	WriteFile("/sys/class/gpio/unexport", value);
}

int FileGpio::Read(int n)
{
	char filename[256];
	snprintf(filename, 256, "/sys/class/gpio/gpio%i/value", n);
	return ReadFile(filename);
}

void FileGpio::Write(int n, bool high)
{
	char filename[256];
	snprintf(filename, 256, "/sys/class/gpio/gpio%i/value", n);
	WriteFile(filename, high ? "1" : "0");
}

int FileGpio::ReadFile(const char *filename)
{
	FILE *f = fopen(filename, "r");
	if (!f){
		fprintf(stderr, "Cannot open file for reading: \"%s\"\n", filename);
		throw 0;
	}
	char c = 0;
	if (1 != fread(&c, 1, 1, f)){
		fprintf(stderr, "Cannot read file: \"%s\"\n", filename);
		throw 0;
	}
	fclose(f);
	return (c != '0');
}

void FileGpio::WriteFile(const char *filename, const char *value)
{
	FILE *f = fopen(filename, "w");
	if (!f){
		fprintf(stderr, "Cannot open file for writing: \"%s\"\n", filename);
		throw 0;
	}
	fprintf(f, "%s", value);
	fclose(f);
}




// Access from ARM Running Linux

#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO controller */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>

#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)


// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))

#define GPIO_SET *(gpio+7)  // sets   bits which are 1 ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0
#define GPIO_VAL *(gpio+13) // values

DirectGpio::DirectGpio()
{
	/* open /dev/mem */
	if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
		fprintf(stderr, "can't open /dev/mem \n");
		exit (-1);
	}

	/* mmap GPIO */

	// Allocate MAP block
	if ((gpio_mem = (char *)malloc(BLOCK_SIZE + (PAGE_SIZE-1))) == NULL) {
		fprintf(stderr, "allocation error \n");
		exit (-1);
	}

	// Make sure pointer is on 4K boundary
	if ((unsigned long)gpio_mem % PAGE_SIZE)
		gpio_mem += PAGE_SIZE - ((unsigned long)gpio_mem % PAGE_SIZE);

		// Now map it
		gpio_map = (char *)mmap(
			(caddr_t)gpio_mem,
			BLOCK_SIZE,
			PROT_READ|PROT_WRITE,
			MAP_SHARED|MAP_FIXED,
			mem_fd,
			GPIO_BASE
		);

	if ((long)gpio_map < 0) {
		fprintf(stderr, "mmap error %d\n", (int)gpio_map);
		exit (-1);
	}

	// Always use volatile pointer!
	gpio = (volatile unsigned *)gpio_map;
}

void DirectGpio::Enable(int n, bool out)
{
	INP_GPIO(n); // must use INP_GPIO before we can use OUT_GPIO
	if (out){
		OUT_GPIO(n);
	}
}

void DirectGpio::Disable(int n)
{
	Write(n, false);
}

int DirectGpio::Read(int n)
{
	return (GPIO_VAL & (1 << n)) != 0;
}

void DirectGpio::Write(int n, bool high)
{
	if (high){
		GPIO_SET = 1 << n;
	}
	else {
		GPIO_CLR = 1 << n;
	}
}

