/*
	rpipic - A PIC Programmer for the Raspberry Pi
	Copyright (C) 2012  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/


#include <cstdio>
#include <cstring>

#include "pic12.hpp"

#define TDLY12 ((m_chip.Tdly1 < m_chip.Tdly2) ? m_chip.Tdly2 : m_chip.Tdly1)

struct Pic12BitChip
{
	const char *name;
	
	int64_t Tppdp;
	int64_t Thld0;
	int64_t Tset1;
	int64_t Thld1;
	int64_t Tdly1;
	int64_t Tdly2;
	int64_t Tdly3;
	int64_t Tera;
	int64_t Tdis;
	int64_t Tprog;
	uint16_t ProgramLength;
	uint16_t DataLength;
};

const Pic12BitChip Pic12Bit::s_chips[] =
{
	//{"pic10f200", 5000, 5000, 100, 100, 1000, 1000, 80, 10000000, 100000, 2000000, 0x100, 0x0},
	{"pic16f54",  5000, 5000, 100, 100, 1000, 1000, 80, 10000000, 100000, 2000000, 0x200, 0x0},
	{0}
};

const Pic12BitChip &Pic12Bit::GetChip(const char *name)
{
	for (unsigned int i = 0; s_chips[i].name != 0; i++){
		if (!strcmp(s_chips[i].name, name)){
			return s_chips[i];
		}
	}
	fprintf(stderr, "Could not find 12 bit chip %s\n", name);
	throw 0;
}

Pic12Bit::Pic12Bit(Board &board, const char *name) :
	Pic(board),
	m_chip(GetChip(name))
{
	InitChip(m_chip.Tdly3, m_chip.Tset1, m_chip.Thld1);
	//-m_chip.Tppdp (rather than +) because 12 bit interface raises vdd then vpp
	InitPic(-m_chip.Tppdp, m_chip.Thld0, TDLY12, m_chip.Tera, m_chip.Tdis, m_chip.Tprog);
	Reset();
}

Pic12Bit::~Pic12Bit()
{
	GetBoard().SetVppHigh(false);
	GetBoard().SetVddHigh(false);
	Wait(100000000);
	GetBoard().SetClockHigh(false);
	Wait(100000000);
	GetBoard().SetDataHigh(false);
	Wait(100000000);
}

void Pic12Bit::PrintMeta(Data &data)
{
}

void Pic12Bit::ReadProgram(Data &data)
{
	fprintf(stderr, "Reading program...\n");
	PicReadProgram(data, 0, 0, m_chip.ProgramLength, 12, false, true);
	fprintf(stderr, "Finished reading program...\n");
}

void Pic12Bit::ReadData(Data &data)
{
	fprintf(stderr, "Reading data...UNIMPLEMENTED\n");
	//PicReadData(data, 0, 0, m_chip.DataLength);
	fprintf(stderr, "Finished reading data...\n");
}

void Pic12Bit::ReadConfiguration(Data &data)
{
	fprintf(stderr, "Reading configuration...\n");
	PicReadProgram(data, 0, 0xfff, 1, 12, false, false);
	fprintf(stderr, "Finished reading configuration...\n");
}

void Pic12Bit::ReadUserId(Data &data)
{
	fprintf(stderr, "Reading user ID... EXPERIMENTAL; UNCOMMENT TO USE\n");
	//PicReadProgram(data, 0, 0x200, 4, 12, false, true);
	fprintf(stderr, "Finished reading user ID...\n");
}

void Pic12Bit::ReadDeviceId(Data &data)
{
	fprintf(stderr, "There is no device ID for 12 bit PICs\n");
	throw 0;
}

void Pic12Bit::WriteProgram(Data &data)
{
	fprintf(stderr, "Programming program...\n");
	PicWriteProgram(data, 0, 0, m_chip.ProgramLength, 12, true, false, true);
	fprintf(stderr, "Finished programming program...\n");
}

void Pic12Bit::WriteData(Data &data)
{
	fprintf(stderr, "Programming data...UNIMPLEMENTED\n");
	//PicWriteData(data, 0, , m_chip->DataLength, false);
	fprintf(stderr, "Finished programming data...\n");
}

void Pic12Bit::WriteConfiguration(Data &data)
{
	fprintf(stderr, "Writing configuration...\n");
	PicWriteProgram(data, 0, 0xfff, 1, 12, true, false, false);
	fprintf(stderr, "Finished writing configuration...\n");
}

void Pic12Bit::WriteUserId(Data &data)
{
	fprintf(stderr, "Reading user ID... EXPERIMENTAL; UNCOMMENT TO USE\n");
	//PicWriteProgram(data, 0, 0x200, 4, 12, true, false, true);
	fprintf(stderr, "Finished writing user ID...\n");
}

void Pic12Bit::Erase()
{
	Reset();
	for (unsigned int i = 0; i < m_chip.ProgramLength; i++){
		IncrementAddress();
	}
	Write(0x09, 6);
	Wait(TDLY12);
	Wait(m_chip.Tera);
}


