/*
	rpipic - A PIC Programmer for the Raspberry Pi
	Copyright (C) 2012  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/


#include <cstdio>

#include "pic14.hpp"

struct Pic14BitChip
{
	const char *name;
	uint16_t id;
	
	int64_t Tppdp;
	int64_t Thld0;
	int64_t Tset1;
	int64_t Thld1;
	int64_t Tdly1;
	int64_t Tdly2;
	int64_t Tdly3;
	int64_t Tera;
	int64_t Tprog;
	uint16_t ProgramLength;
	uint16_t DataLength;
};

#define TDLY12 ((m_chip->Tdly1 < m_chip->Tdly2) ? m_chip->Tdly2 : m_chip->Tdly1)
const Pic14BitChip Pic14Bit::s_chips[] =
{
	{"pic16f627", 0x07A0, 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, 0x400, 0x80},
	{"pic16f628", 0x07C0, 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, 0x800, 0x80},
	
	//{"pic16f870", , 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, , 64},
	//{"pic16f871", , 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, , 64},
	{"pic16f872", 0x08E0, 5000, 5000, 100, 100, 1000, 1000, 80, 4000000, 4000000, 2048, 64},
	//{"pic16f873", , 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, , 128},
	//{"pic16f874", , 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, , 128},
	//{"pic16f876", , 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, , 256},
	//{"pic16f877", , 5000, 5000, 100, 100, 1000, 1000, 80, 5000000, 8000000, , 256},
	
	{0}
};

const Pic14BitChip *Pic14Bit::GetChip(uint16_t word, bool error)
{
	for (unsigned int i = 0; s_chips[i].name != 0; i++){
		if ((0x3FE0 & word) == s_chips[i].id){
			return &s_chips[i];
		}
	}
	if (error){
		fprintf(stderr, "Could not find 14 bit chip with ID: 0x%04X\n", word);
	}
	throw 0;
}

Pic14Bit::Pic14Bit(Board &board) :
	Pic(board),
	m_chip(GetChip(GetDeviceId(), true))
{
	fprintf(stderr, "Detected chip: %s\n", m_chip->name);
	
	InitChip(m_chip->Tdly3, m_chip->Tset1, m_chip->Thld1);
	InitPic(m_chip->Tppdp, m_chip->Thld0, TDLY12, m_chip->Tera, 0, m_chip->Tprog);
	
	Reset();
}

void Pic14Bit::ReadProgram(Data &data)
{
	fprintf(stderr, "Reading program...\n");
	PicReadProgram(data, 0, 0, m_chip->ProgramLength, 14);
	fprintf(stderr, "Finished reading program...\n");
}

void Pic14Bit::ReadData(Data &data)
{
	fprintf(stderr, "Reading data...\n");
	PicReadData(data, 0, 0x2100, m_chip->DataLength);
	fprintf(stderr, "Finished reading data...\n");
}

void Pic14Bit::ReadConfiguration(Data &data)
{
	fprintf(stderr, "Reading configuration...\n");
	PicReadProgram(data, 7, 0x2007, 1, 14, true);
	fprintf(stderr, "Finished reading configuration...\n");
}

void Pic14Bit::ReadUserId(Data &data)
{
	fprintf(stderr, "Reading user ID...\n");
	PicReadProgram(data, 0, 0x2000, 4, 14, true);
	fprintf(stderr, "Finished user ID...\n");
}

void Pic14Bit::ReadDeviceId(Data &data)
{
	uint16_t word = GetDeviceId();
	if (0x3FFF != word){ //assume the default is all 1s
		data.SetData(0x2006 * 2, word & 0xFF);
		data.SetData(0x2006 * 2 + 1, word >> 8);
	}
}

uint16_t Pic14Bit::GetDeviceId()
{
	fprintf(stderr, "Reading device ID...\n");
	Reset();
	LoadConfiguration();
	for (unsigned int i = 0; i < 6; i++){
		IncrementAddress();
	}
	uint16_t word = ReadProgramWord(14);
	fprintf(stderr, "Finished reading device ID...\n");
	return word;
}

void Pic14Bit::WriteProgram(Data &data)
{
	fprintf(stderr, "Programming program...\n");
	PicWriteProgram(data, 0, 0, m_chip->ProgramLength, 14, false);
	fprintf(stderr, "Finished programming program...\n");
}

void Pic14Bit::WriteData(Data &data)
{
	fprintf(stderr, "Programming data...\n");
	PicWriteData(data, 0, 0x2100, m_chip->DataLength, false);
	fprintf(stderr, "Finished programming data...\n");
}

void Pic14Bit::WriteConfiguration(Data &data)
{
	fprintf(stderr, "Programming configuration...\n");
	PicWriteProgram(data, 7, 0x2007, 1, 14, false, true);
	fprintf(stderr, "Finished programming configuration...\n");
}

void Pic14Bit::WriteUserId(Data &data)
{
	fprintf(stderr, "Programming user ID...\n");
	PicWriteProgram(data, 0, 0x2000, 4, 14, false, true);
	fprintf(stderr, "Finished programming user ID...\n");
}

void Pic14Bit::Erase()
{
	fprintf(stderr, "Erasing PIC...\n");
	Reset();
	LoadConfiguration();
	for (unsigned int i = 0; i < 7; i++){
		IncrementAddress();
	}
	WriteProgramWord(0x0000, 14);
	BeginEraseProgramCycle();
	
	Reset();
	LoadConfiguration();
	for (unsigned int i = 0; i < 7; i++){
		IncrementAddress();
	}
	BulkEraseSetup();
	WriteProgramWord(0x3FFF, 14);
	BeginEraseProgramCycle();
	Wait(TDLY12);
	BulkEraseSetup();
	
	Reset();
	fprintf(stderr, "Finished erasing PIC...\n");
}

void Pic14Bit::BulkEraseSetup()
{
	Write(0x1, 6);
	Wait(m_chip->Tdly2);
	Wait(TDLY12);
	
	Write(0x7, 6);
	Wait(m_chip->Tdly2);
	Wait(TDLY12);
}

void Pic14Bit::PrintMeta(Data &data)
{
	for (unsigned int i = 0; i < 4; i++){
		try {
			uint16_t word = data.GetData((0x2000 + i) * 2) | (data.GetData((0x2000 + i) * 2 + 1) << 8);
			printf("User ID %u: 0x%04x  %s\n", i, word, FormatBits(word, 14));
		}
		catch(int){}
	}
	
	try {
		uint16_t word = data.GetData(0x2006 * 2) | (data.GetData(0x2000 * 2 + 1) << 8);
		try {
			const Pic14BitChip *chip = GetChip(word, false);
			printf("Device ID: %s  0x%04x  %s\n", chip->name, word, FormatBits(word, 14));
		}
		catch (int) {
			printf("Device ID: UNKNOWN    0x%04x  %s\n", word, FormatBits(word, 14));
		}
	}
	catch(int){}
	
	try {
		uint16_t word = data.GetData(0x2007 * 2) | (data.GetData(0x2007 * 2 + 1) << 8);
		printf("Configuration Word: 0x%04x  %s\n", word, FormatBits(word, 14));
		switch (((word & 0x0010) >> 2) | (word & 0x0003))
		{
			case 0x00:
				printf("\tOscillator: LP (Low power on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)\n");
				break;
			case 0x01:
				printf("\tOscillator: XT (Crystal/resonator on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)\n");
				break;
			case 0x02:
				printf("\tOscillator: HS (High speed crystal/resonator on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)\n");
				break;
			case 0x03:
				printf("\tOscillator: ExtClk (IO function on RA6/OSC2/CLKOUT, CLKIN on RA7/OSC1/CLKIN)\n");
				break;
			case 0x04:
				printf("\tOscillator: INTRC (IO function on RA6/OSC2/CLKOUT, IO function on RA7/OSC1/CLKIN)\n");
				break;
			case 0x05:
				printf("\tOscillator: INTRC (CLKOUT function on RA6/OSC2/CLKOUT, IO function on RA7/OSC1/CLKIN)\n");
				break;
			case 0x06:
				printf("\tOscillator: ER (IO function on RA6/OSC2/CLKOUT, Resistor on RA7/OSC1/CLKIN)\n");
				break;
			case 0x07:
				printf("\tOscillator: ER (CLKOUT function on RA6/OSC2/CLKOUT, Resistor function on RA7/OSC1/CLKIN)\n");
				break;
			default:
				printf("\tOscillator: Unknown\n");
		}
	
		printf("\tWatchdog Timer: %s\n", ((word >> 2) & 1) ? "Enabled" : "Disabled");
		printf("\tPower up timer: %s\n", ((word >> 3) & 1) ? "Disabled" : "Enabled");
		printf("\tMCLR: RA5/MCLR function is %s\n", ((word >> 5) & 1) ? "MCLR" : "digital input (MCLR internally tied to Vdd)");
		printf("\tBrown Out Detect: %s\n", ((word >> 6) & 1) ? "Enabled" : "Disabled");
		printf("\tLow Voltage Programming: %s\n", ((word >> 7) & 1) ? "Enabled" : "Disabled");
		printf("\tData Protection: %s\n", ((word >> 8) & 1) ? "Disabled" : "Enabled");
	
		switch ((word >> 10 & 0xF))
		{
			case 0x0:
				printf("\tCode Protection: 0x0000-0x03FF code protected\n");
				break;
			case 0x5:
				printf("\tCode Protection: 0x0200-0x03FF code protected\n");
				break;
			case 0xA:
				printf("\tCode Protection: 0x0400-0x03FF code protected (or disabled)\n");
				break;
			case 0xF:
				printf("\tCode Protection: Disabled\n");
				break;
			default:
				printf("\tCode Protection: Unset\n");
		}
	}
	catch(int){}
}

